import io
import base64
import requests
import os
import warnings

# noinspection PyPackageRequirements
import ipywidgets as widgets

from collections import defaultdict, OrderedDict
from contextlib import redirect_stdout

from dli import connect
from ipywidgets import Layout, HTML
from IPython.core.magic import (Magics, magics_class, line_magic)  # noqa: I900
from IPython import get_ipython  # noqa: I900
from IPython.display import Javascript, display  # noqa: I900


warnings.filterwarnings('ignore')


def help(obj):
    print("This is a " + type(obj).__name__ + "\n" + "*"*20,
        "\nIt has these functions:\n" + "\n".join(
          ["." + method_name for method_name in dir(obj)
                  if callable(getattr(obj, method_name))
           and ((not method_name.startswith("_")) or method_name=='__iter__')])
        ,
        "\n\nIt has these properties:\n" + "\n".join(
          ["." + method_name for method_name in dir(obj)
                  if not callable(getattr(obj, method_name))
           and ((not method_name.startswith("_")) or method_name=='__iter__')])
    + "\n\nAnd it looks like this:\n" + str(obj)
         )


def redirect(fun):
    with io.StringIO() as buf, redirect_stdout(buf):
        fun()
        output = buf.getvalue()
        return output


def embedded_image(url):
    response = requests.get(url)
    uri = ("data:" +
           response.headers['Content-Type'] + ";" +
           "base64," + str(
                base64.b64encode(response.content).decode('utf-8')))
    return uri


class mockRefresh():

        def __init__(self, msg="Refreshing"):
            self.msg = msg

        def __str__(self):
            return self.msg

        def __repr__(self):
            return self.msg

        def datasets(self):
            return []

        def contents(self):
            return ""

        def shape(self):
            return ""

        def description(self):
            return ""

        @property
        def instances(self):
            class mockInstance():

                def all(self):
                    return []

            return mockInstance()


class SelectorWidget:
    def __init__(self, ipython):
        #  self.evaluate_login()

        self.ipython = ipython
        self._name_counts = defaultdict(lambda: 0)
        self.packages = []
        self.datasets = []
        self.current_obj = None
        self.dl = None
        self.mockrefresh = mockRefresh()
        self.mocknoitems = mockRefresh("No entries")

        html = f'<img src="{embedded_image("https://cdn.ihsmarkit.com/www2/a/p/media/images/ihsmarkit.svg")}" />'
        self.imagery = HTML(html, layout=Layout(width='20%', height='60px'))
        self.env_map = {
            '1':'https://catalogue.datalake.ihsmarkit.com/__api',
            '2':'https://catalogue-uat.datalake.ihsmarkit.com/__api',
            '3':'https://catalogue-qa.udpmarkit.net/__api'
        }

        # Widgets ---------------------------------------------------------
        self.api_key = widgets.Text(
            value='',
            placeholder='Enter your api key',
            description='Api key',
            disabled=False,
            layout=Layout(width='100%', height='30px'),
            style={'border-radius': '5px'}
        )

        self.env_dd = widgets.Dropdown(
            options=[('Prod', 1), ('UAT', 2), ('QA', 3)],
            value=1,
            description='Environment',
            disabled=False,
            layout=Layout(width='100%'),
            style={'border-radius': '5px'}
        )

        self.package_dd = widgets.Select(
            options=[],
            description='Packages',
            disabled=False,
            layout=Layout(width='99%'),
            style={'border-radius': '5px'}
        )

        self.dataset_dd = widgets.Select(
            options=[],
            description='Datasets',
            disabled=False,
            layout=Layout(width='99%')
        )

        self.search_package = widgets.Text(
            value='',
            placeholder='Search Package',
            description='Search:',
            disabled=False,
            style={'border-radius': '5px'},
            layout=Layout(width='99%')
        )

        self.search_dataset = widgets.Text(
            value='',
            placeholder='Search Dataset',
            description='Search:',
            disabled=False,
            style={'border-radius': '5px'},
            layout=Layout(width='99%')
        )

        # Buttons ---------------------------------------------------------
        self.connect_button = widgets.Button(
            description='Connect',
            disabled=False,
            button_style='',
            tooltip='Click me',
            icon='check',
            style={'border-radius': '5px'}
        )

        self.to_api_button = widgets.Button(
            description='switch login...',
            disabled=False,
            button_style='',
            tooltip='to api login',
            # icon='check',
            style={'border-radius': '5px'}
        )

        self.to_ccred_button = widgets.Button(
            description='switch login...',
            disabled=False,
            button_style='',
            tooltip='to client credentials login',
            # icon='check',
            style={'border-radius': '5px'}
        )

        self.copy_button = widgets.Button(
            description='Copy current object',
            disabled=False,
            button_style='',
            tooltip='Click me',
            icon='check',
            layout=Layout(width='100%', height='30px')
        )

        # Dropdowns ---------------------------------------------------------
        self.which_login = widgets.Dropdown(
            options=[('Api key', 1),
                     ('Single sign-on', 2),
                     ('Client credentials', 3)
                     ],
            value=1,
            button_style='',
            tooltip='to client credentials login',
            icon='check',
            disabled=False,
            layout=Layout(width='15%'),
            style={'border-radius': '5px'}
        )

        self.which_packages = widgets.ToggleButtons(
            options=['Only Mine', 'All Packages'],
            disabled=False,
            button_style='',
            style={'border-radius': '5px'},
            layout=Layout(width='100%', height='10%'),
            description='View',
        )

        # Handlers ---------------------------------------------------------
        self.connect_button.on_click(self.connect_button_handler)
        self.copy_button.on_click(self.copy_button_handler)

        # Observers ---------------------------------------------------------
        self.search_package.observe(self.on_search_package)
        self.search_dataset.observe(self.on_search_dataset)
        self.package_dd.observe(self.on_change_package)
        self.dataset_dd.observe(self.on_change_dataset)
        self.which_packages.observe(self.on_change_view)

        # Search --------------------------------------------------------
        items = [
            widgets.VBox([self.search_package, self.package_dd], layout=Layout(width='50%')),
            widgets.VBox([self.search_dataset, self.dataset_dd], layout=Layout(width='50%'))
        ]

        # Results -------------------------------------------------------
        tab_contents = ['.contents()', '.shape', '.description', 'print', '__iter__']
        children = [widgets.Textarea(
            value='',
            placeholder='',
            disabled=False,
            layout=Layout(width='100%', height='180px')
        ) for name in tab_contents]

        self.tab = widgets.Tab(
            layout=Layout(width='100%', height='250px'),
        )

        self.tab.children = children
        [self.tab.set_title(num, name) for num, name in enumerate(tab_contents)]

        vbox_logo = widgets.VBox([self.imagery])
        login = [widgets.HBox([self.api_key]),
                 widgets.HBox([self.env_dd,self.which_login, self.connect_button])]
        vbox1 = widgets.VBox(login)
        vbox2 = widgets.HBox(items)
        vbox3 = widgets.HBox([self.which_packages], layout=Layout(width='100%', height='60px'))
        vbox4 = widgets.HBox([self.tab], layout=Layout(width='100%'))

        self.root = widgets.VBox([vbox_logo, vbox1, vbox2, vbox3, vbox4, self.copy_button])


    def _generate_name(self, type_):
        while True:
            self._name_counts[type_] = (
                self._name_counts[type_] + 1
            )

            name = '{}_{}'.format(
                type_, self._name_counts[type_]
            )

            if name not in self.ipython.user_global_ns:
                break
            else:
                return self._generate_name(type_)

        return name

    def copy_button_handler(self, obj):
        _name_alias = {
            'PackageModel': 'package',
            'DatasetModel': 'dataset',
        }
        if (type(self.current_obj).__name__ is not 'PackageModel' and
            type(self.current_obj).__name__ is not 'DatasetModel'):
            print('Cannot retrieve object currently.')
        else:
            name = self._generate_name(
                _name_alias[type(self.current_obj).__name__]
            )
            self.ipython.user_global_ns[name] = self.current_obj
            encoded_code = base64.b64encode(name.encode()).decode()
            display(Javascript("""
                var code = IPython.notebook.insert_cell_{0}('code');
                code.set_text(atob("{1}"));
            """.format('below', encoded_code)))

    def on_change_view(self, change):
        if change['type'] == 'change' and change['name'] == 'value':
            self.package_dd.options = [self.mockrefresh]
            self.dataset_dd.options = []

            if change["new"] == 'All Packages':
                self.packages = self.dl.packages()
            else:
                self.packages = self.dl.packages(only_mine=True)

            self.package_dd.options = self.packages
            if len(self.packages.keys()) > 0:
                self.datasets = self.packages[
                    list(self.packages.keys())[0]].datasets()
                self.dataset_dd.options = self.datasets
            else:
                self.package_dd.options = [self.mocknoitems]
                self.dataset_dd.options = [self.mocknoitems]

    def start_api_session(self):
        self.dl = connect(
            self.api_key.value,
            root_url=self.env_map[str(self.env_dd.value)]
        )

    def start_clientcred_session(self):
        user = os.environ.get("DLI_ACCESS_KEY_ID")
        pasw = os.environ.get("DLI_SECRET_ACCESS_KEY")
        if user is not None and pasw is not None:
            self.dl = connect(root_url=self.env_map[str(self.env_dd.value)])
        else:
            print(f'Set DLI_ACCESS_KEY_ID and DLI_SECRET_ACCESS_KEY '
                  f'environment variables for client credentials login, '
                  f'alternatively use the api/web flow.')

    def start_web_session(self):
        self.dl = connect(root_url=self.env_map[str(self.env_dd.value)])

    def connect_button_handler(self, obj):
        login_options = {'1': self.start_api_session,
                         '2': self.start_web_session,
                         '3': self.start_clientcred_session,
                         }
        try:
            login_options[str(self.which_login.value)].__call__()
        except Exception as e:
            print(f'Session was not started. {e}')

        if self.dl is not None:
            self.packages = self.dl.packages()
            self.package_dd.options = self.packages
            if len(self.packages.keys()) > 0:
                self.datasets = self.packages[list(self.packages.keys())[0]].datasets()
                self.dataset_dd.options = self.datasets
            else:
                self.package_dd.options = [self.mocknoitems]
                self.dataset_dd.options = [self.mocknoitems]

    def on_change_package(self, change):
        if change['type'] == 'change' and change['name'] == 'value':
     
            self.current_obj = change["new"]
            if self.current_obj is not None:
                self.dataset_dd.options = self.current_obj.datasets()
            #current_obj will be overwritten
            self.current_obj = change["new"]
            if self.current_obj:
                self.tab.children[0].value = redirect(self.current_obj.contents)
                self.tab.children[1].value = str(self.current_obj.shape)
                self.tab.children[2].value = str(self.current_obj.description)
                self.tab.children[3].value = str(self.current_obj.__str__())
                self.tab.children[4].value = "\n".join([str(x) for x in
                                                   self.current_obj.datasets()])

    def on_change_dataset(self, change):
        if change['type'] == 'change' and change['name'] == 'value':
            self.current_obj = change["new"]
            if self.current_obj:
                self.tab.children[0].value = redirect(self.current_obj.contents)
                self.tab.children[1].value = ""
                self.tab.children[2].value = ""
                self.tab.children[3].value = str(self.current_obj.__str__())
                self.tab.children[4].value = "\n".join([str(x) for x in
                                                   self.current_obj.instances.all()])

    def on_search_package(self, change):
        if change['type'] == 'change' and change['name'] == 'value':
            self.current_obj = change["new"]
            if self.current_obj:
                lst = list(filter((lambda x: self.current_obj.lower() in x[0].lower()),
                                  self.packages.items()))
                if len(lst) > 0:
                    try:
                        self.package_dd.options = OrderedDict(lst)
                        self.dataset_dd.options = lst[0][1].datasets()
                    except Exception as e:
                        print(e)
                else:
                    self.package_dd.options = [self.mocknoitems]
                    self.dataset_dd.options = [self.mocknoitems]
            else:
                self.package_dd.options = self.packages
                first = self.packages[list(self.packages.keys())[0]]
                self.dataset_dd.options = first.datasets()

    def on_search_dataset(self, change):
        if change['type'] == 'change' and change['name'] == 'value':
            self.current_obj = change["new"]
            if self.current_obj:
                lst = list(filter((lambda x: self.current_obj.lower() in x[0].lower()),
                                  self.datasets))
                if len(lst) > 0:
                    self.dataset_dd.options = OrderedDict(lst)
                else:
                    self.dataset_dd.options = [self.mocknoitems]
            else:
                self.dataset_dd.options = self.datasets


@magics_class
class DatasetSelectorMagic(Magics):

    @line_magic
    def dataset_selector(self, line):
        disp = SelectorWidget(self.parent)
        self.parent.user_global_ns['_widget'] = disp
        display(disp.root)


def load_ipython_extension(ipython):
    # The `ipython` argument is the currently active `InteractiveShell`
    # instance, which can be used in any way. This allows you to register
    # new magics or aliases, for example.
    ipython.register_magics(DatasetSelectorMagic)


def unload_ipython_extension(ipython):
    # If you want your extension to be unloadable, put that logic here.
    pass


test = {}


def show():
    get_ipython().magic('dataset_selector')


def launch():
    get_ipython().extension_manager.load_extension('dli.gui')
    show()


# if __name__ != '__main__':
#     launch()


import re
import warnings
from collections import OrderedDict
from typing import Dict

from dli.client.exceptions import CatalogueEntityNotFoundException
from dli.models.paginator import Paginator
from dli.client.components.urls import package_urls
from dli.models.package_model import PackageModel


class PackageModule:

    @staticmethod
    def _search_terms_to_filters(search_terms, start_count):

        def to_dict(k, op, v):
            return {'key': k.strip(), 'op': op.strip(), 'val': v.strip()}

        def partition_search_term(x):
            split = re.split("(<|>|=>|<=|=|!=| contains | like | ilike )", x)
            if len(split) == 3:
                return to_dict(*split)
            else:
                return None

        arguments = map(partition_search_term, search_terms)
        enumerated_partitions = [
            (x, y) for x, y in enumerate(arguments, start=start_count)
            if y is not None
        ]
        opmp = {
            "=": "eq",
            ">=": "gte",
            "<=": "lte",
            ">": "gt",
            "<": "lt",
        }

        wildcard = lambda op, val:  \
            "%" + val + "%" if op in ["like", "ilike"] else val

        to_numbered_filter = lambda partition: (lambda idx, x: {
            f'filter[{idx}][field]': x['key'],
            f'filter[{idx}][operator]': opmp.get(x['op'], x['op']),
            f'filter[{idx}][value]': wildcard(x['op'], x['val'])

        })(*partition)

        yield from map(to_numbered_filter, enumerated_partitions)

    @staticmethod
    def _filter_creation(search_term, only_mine):
        bump = 0
        if only_mine:
            yield {
                'filter[0][value]': True,
                'filter[0][operator]': 'eq',
                'filter[0][field]': 'has_access',
            }
            bump = 1
        else:
            yield {}

        if type(search_term) is str:
            search_term = [search_term]

        if search_term and type(search_term) is list:
            yield from PackageModule._search_terms_to_filters(search_term, bump)

    def __call__(self, search_term=None, only_mine=True) \
            -> Dict[str, PackageModel]:
        """
        See packages we can access at the top level.

        :param bool only_mine: Specify whether to collect packages only
        accessible to you (user) or to discover packages that you may
        want to discover.

        :returns: Ordered dictionary of id to PackageModel.
        :rtype: OrderedDict[id: str, PackageModel]
        """

        # search_term = ["y=z", "a>c", "d", "f contains pull", "f ilike pig"]
        filters = {}
        for x in PackageModule._filter_creation(search_term, only_mine):
            filters.update(x)

        search_paginator = Paginator(
            package_urls.v2_package_index,
            self._client._Package,
            self._client._Package._from_v2_response_unsheathed,
            page_size=5000,
            filters=filters
        )

        return OrderedDict([(v.name, v) for v in search_paginator])

    def get(self, name) -> PackageModel:
        """
        Find a PackageModel with the matching name. If not found then
        returns None.

        :param str name: The name of the package to collect
        (short codes are yet to be implemented for packages)
        :returns: PackageModel with matching name.
        :rtype: PackageModel
        """
        warnings.warn(
            'Getting a package by name'
            'will be deprecated in future. Short-codes will replace this.',
            PendingDeprecationWarning
        )

        res = self._client.packages(search_term=[f"name={name}"]).get(name)
        if res:
            return res
        else:
            raise Exception(f"No such package {name}")



from collections import OrderedDict
from typing import Dict

from dli.client.exceptions import CatalogueEntityNotFoundException
from dli.models.paginator import Paginator
from dli.client.components.urls import dataset_urls, identity_urls
from dli.models.dataset_model import DatasetModel
from dli.modules.package_module import PackageModule


class DatasetModule:

    def __call__(self, search_term=None, only_mine=True) \
            -> Dict[str, DatasetModel]:
        """
        See datasets.

        :param bool only_mine: Specify whether to collect datasets only
        accessible to you (True) or to discover packages that you may
        want to discover but may not have access to (False).

        :returns: Ordered dictionary of ids to DatasetModel.
        :rtype: OrderedDict[id: str, DatasetModel]
        """

        filters = {}
        for x in PackageModule._filter_creation(search_term, only_mine):
            filters.update(x)

        p = Paginator(
            dataset_urls.v2_index,
            self._client.Dataset,
            self._client.Dataset._from_v2_response_unsheathed,
            page_size=5000,
            max_workers=5,
            filters=filters
        )

        results = [(v.short_code, v) for v in p]
        o = OrderedDict()
        for code, val in results:
            if code in o:
                o[code] = [o[code]]
                o[code].append(val)
            else:
                o[code] = val

        return o

    def get(self, short_code, organisation_short_code=None) -> DatasetModel:
        """
        Returns a DatasetModel if it exists, and a user has access else None

        :param str short_code: The short code of the dataset to collect

        :returns: Dataset model with matching short code.
        :rtype: DatasetModel
        """

        search_terms = [
            f"short_code={short_code}"
        ]

        if organisation_short_code is not None:
            search_terms.append(
                f"organisation_short_code={organisation_short_code}"
            )

        def org_id_to_shortcode(organisation_id: str):
            try:
                org = self._client._Organisation._from_v2_response(self._client.session.get(
                    identity_urls.org_by_id.format(id=organisation_id)
                ).json())
                return org
            except Exception as e:
                raise(e)

        # todo - not going to work as expected since the orderdict wont have dups
        res = self._client.datasets(search_term=search_terms).get(short_code)

        if type(res) is list:
            org_codes=[f"\033[4m\033[92m\033[1morganisation_short_code={org_id_to_shortcode(x.organisation_id).short_code}\033[0;0m -> {x.short_code}"
                       for x in res if x.has_access]
            msg = f"\033[96mMultiple datasets have shortcode: \033[1m`{short_code}`\033[0;0m\n\nShortcodes are"\
                  f" only guaranteed to be unique per organisation, and you seem to "\
                  f" have access to multiple organisation's datasets, where there are "\
                  f"datasets bearing the same shortcode."\
                  f"\n"\
                  f"To resolve this, please specify the organisation_short_code in "\
                  f"the call to datasets.get(shortcode, organisation_short_code=ORG CODE)"\
                  f"\n\nThe available (accessible to you) organisations shortcodes "\
                  f"for this dataset short code are:\n" + ("\n".join(org_codes))

            raise Exception(
                msg
            )

            #todo - difficult here, really need to convert id to shortcode for org
            #todo - else isnt that useful a message as user now needs to discover these
        elif res:
            return res
        else:
            raise Exception(f"No such dataset {short_code}")


import operator
import re
from collections import OrderedDict

import pytest

from dli.client.dli_client import DliClient
from datetime import datetime
from unittest.mock import MagicMock


class TestClientMock(DliClient):

    def __init__(self, args, kwargs):
        self._session = MagicMock()
        self._session.auth_key = 'abc'
        self._session.token_expires_on = datetime(2100, 1, 1)
        self._environment = MagicMock()
        self._environment.consumption = ''
        self._environment.s3_proxy = 's3proxy.fake'
        self._analytics_handler = MagicMock()
        self.packages = self._packages()
        self.datasets = self._datasets()

    @property
    def session(self):
        return self._session


@pytest.fixture
def test_client():
    yield TestClientMock('abc', '123')


def mock_catalogue_filter_search_response_impl(body, args, **kwargs):
    # very rudimentary impl of a catalogue parsing the request
    # and responding appropriately given a corpus of packages

    # we're `effectively` testing against a mock here - but otherwise we
    # could only rely on the response of the fixture which we chose
    # or as full-blown integration, so this serves more as a documentation
    # of what catalogue must be doing with our request.

    res = {}
    for x, y in kwargs['params'].items():
        # remove redundant keys
        if x.startswith('filter'):
            try:
                res[re.search(r"\[(\w+)\]\[(\w+)\]", x).groups()] = y
            except Exception as e:
                raise e

    groups = OrderedDict()
    for x in (range(0, int(len(res.keys()) / 3))):
        for key, val in res.items():
            # we'll handle only 0-9 in the tests else this falls down
            if key[0] == str(x):
                if not x in groups:
                    groups[x] = []

                groups[x].append((key[1], val))

    candidates = []
    get_op = {
        "eq": operator.eq,
        "gte": operator.ge,
        "lte": operator.le,
        "gt": operator.gt,
        "lt": operator.lt,
        "like": str.__contains__,
        "ilike": str.__contains__,
        "contains": str.__contains__
    }

    for item in body["data"]:
        passing = True
        for x in list(groups.values()):
            d = dict(x)
            if d['field'] == 'id':
                item_attrib = item.get("id")
            else:
                item_attrib = item["attributes"].get(d['field'])
            like_op = d['operator'] in ['like', 'ilike', 'contain']
            op = get_op[d['operator']]
            if like_op \
                    and not (d['value'].replace("%", "") in item_attrib):
                passing = False
                break
            elif not like_op and not op(item_attrib, d['value']):
                passing = False
                break

        if passing:
            candidates.append(item)

    class Temp():

        def __init__(self, candidates):
            self.candidates = candidates

        def json(self):
            return {'data': self.candidates, 'meta': body["meta"]}

    return Temp(candidates)

@pytest.fixture
def client():
    yield TestClientMock('abc', '123')



@pytest.fixture
def package_request_index_accessible_content_v2():
    return {
        'type': 'package',
        'attributes': {
            'is_internal_within_organisation': True,
            'topic': 'Climate',
            'keywords': [],
            'updated_by': 'api:datalake-mgmt:Timothy.Ryles@ihsmarkit.com',
            'access_manager_id': 'datalake-mgmt',
            'created_by': 'api:datalake-mgmt:Timothy.Ryles@ihsmarkit.com',
            'organisation_name': 'IHS Markit',
            'organisation_id': '9516c0ba-ba7e-11e9-8b34-000c6c0a981f',
            'description': 'test-data',
            'name': 'Test Package',
            'tech_data_ops_id': 'datalake-mgmt',
            'updated_at': '2020-05-13T15:08:07.252878',
            'access': 'Restricted',
            'publisher': 'datalake-mgmt',
            'terms_and_conditions': 'Terms',
            'content_types': ['Reference Data'],
            'has_access': True,
            'data_sensitivity': 'Public',
            'taxonomy': ['Data-IHS Markit-Climate'],
            'contract_ids': [],
            'created_at': '2020-05-13T15:08:07.297885',
            'manager_id': 'datalake-mgmt'
        },
        'id': '8d2781c2-952b-11ea-892f-aa2778cbc602'
    }


@pytest.fixture()
def package_request_index_unaccessible_content_v2():
    return {
        'type': 'package',
        'attributes': {
            'is_internal_within_organisation': True,
            'topic': 'Finance',
            'keywords': [],
            'updated_by': 'api:datalake-mgmt:Timothy.Ryles@ihsmarkit.com',
            'access_manager_id': 'datalake-mgmt',
            'created_by': 'api:datalake-mgmt:Timothy.Ryles@ihsmarkit.com',
            'organisation_name': 'Test Org',
            'organisation_id': '9516c0ba-ba7e-11e9-8b34-000c6c0a981f',
            'description': 'test-data',
            'name': 'Different Name',
            'tech_data_ops_id': 'datalake-mgmt',
            'updated_at': '2020-05-13T15:08:07.252878',
            'access': 'Restricted',
            'publisher': 'datalake-mgmt',
            'terms_and_conditions': 'Terms',
            'content_types': ['Reference Data'],
            'has_access': False,
            'data_sensitivity': 'Public',
            'taxonomy': ['Data-Test Org-Finance'],
            'contract_ids': [],
            'created_at': '2020-05-13T15:08:07.297885',
            'manager_id': 'datalake-mgmt'
        },
        'id': '8d2781c2-952b-11ea-892f-aa2778cbc602'
    }


@pytest.fixture
def package_request_index_v2(package_request_index_accessible_content_v2,
                             package_request_index_unaccessible_content_v2):
    return {
        'data': [package_request_index_accessible_content_v2, package_request_index_unaccessible_content_v2],
        'meta': {
           'total_count': 2, 'page': 1, 'page_size': 5000, 'total_pages': 1
        }
    }

@pytest.fixture
def dataset_request_index_accessible_content_v2():
    return {
        'type': 'dataset',
         'attributes': {
             'package_id': '59e6a698-975d-11e9-84cc-7e5ef76d533f',
             'updated_by': 'update@ihsmarkit.com',
             'keywords': [],
             'location': {},
             'has_access': True,
             'first_datafile_at': '2019-05-06',
             'short_code': 'TestDataset',
             'absolute_taxonomy': 'Data-Test',
             'publishing_frequency': 'Monthly',
             'name': 'Test DataSet',
             'last_datafile_at': '2020-01-01',
             'data_format': 'PARQUET',
             'created_at': '2019-06-25',
             'description': 'description',
             'organisation_name': 'IHS Markit',
             'created_by': 'creator@ihsmarkit.com',
             'taxonomy': []
         },
         'id': '5b01376e-975d-11e9-8832-7e5ef76d533f'
    }
@pytest.fixture
def dataset_request_index_unaccessible_content_v2():
    return {
        'type': 'dataset',
         'attributes': {
             'package_id': '59e6a698-975d-11e9-84cc-0000000000',
             'updated_by': 'update@ihsmarkit.com',
             'keywords': [],
             'location': {},
             'has_access': False,
             'first_datafile_at': '2019-05-06',
             'short_code': 'OtherDataset',
             'absolute_taxonomy': 'Data-Test',
             'publishing_frequency': 'Monthly',
             'name': 'Other DataSet',
             'last_datafile_at': '2020-01-01',
             'data_format': 'PARQUET',
             'created_at': '2019-06-25',
             'description': 'description',
             'organisation_name': 'IHS Markit',
             'created_by': 'creator@ihsmarkit.com',
             'taxonomy': []
         },
         'id': '5b01376e-975d-11e9-8832-7e5ef76d533f'
    }

@pytest.fixture
def dataset_request_index_v2(dataset_request_index_accessible_content_v2,
                             dataset_request_index_unaccessible_content_v2):
    return {
            'data':
                [dataset_request_index_accessible_content_v2,
                 dataset_request_index_unaccessible_content_v2
                ],
            'meta': {'total_count': 2, 'page': 1,
                     'total_pages': 1}
          }


@pytest.fixture
def package_request_v2():
    return {
        'data': {
            'id': '1234',
            'attributes': {
                'name': 'Test Package',
                'topic': 'Test Topic',
                'keywords': ['test'],
                'hasAccess': True,
                'description': 'DESC',
                'documentation': 'href',
                'dataSensitivity': 'Private',
            }
        }
    }


@pytest.fixture
def dataset_request_v2():
    return {
            'data': {
                'type': 'dataset',
                'attributes': {
                     'package_id': '59e6a698-975d-11e9-84cc-7e5ef76d533f',
                     'updated_by': 'update@ihsmarkit.com',
                     'keywords': [],
                     'location': {},
                     'has_access': False,
                     'first_datafile_at': '2019-05-06',
                     'short_code': 'TestDataset',
                     'absolute_taxonomy': 'Data-Test',
                     'publishing_frequency': 'Monthly',
                     'name': 'Test DataSet',
                     'last_datafile_at': '2020-01-01',
                     'data_format': 'PARQUET',
                     'created_at': '2019-06-25',
                     'description': 'description',
                     'organisation_name': 'IHS Markit',
                     'created_by': 'creator@ihsmarkit.com',
                     'taxonomy': []
                 },
             'id': '5b01376e-975d-11e9-8832-7e5ef76d533f'
        }
      }


@pytest.fixture
def instance_request_v1():
    return {
        'properties': {'pages_count': 1},
        'entities': [{
            'properties': {'datafileId': 1,  'total_size': 1048576}
        }]
    }


@pytest.fixture
def dictionary_request_v2():
    return {
        'data': {
        'type': 'dictionary',
        'attributes': {
          'dataset_id': '7eb6eb49-5fae-44bc-900a-c241e0a24b00',
          'description': None,
          'version': '1',
          'partitions': [],
          'valid_as_of': '2019-09-26'
        },
        'id': '2d113c3a-be1a-4c70-b87a-8e569dee8dd0'}
    }


@pytest.fixture
def fields_request_v2():
    return {
        'data':
        {
            'type': 'dictionary_fields',
            'attributes': {
                  'fields': [
                    {'nullable': True,
                     'type': 'String',
                     'description': 'IHS',
                     'name': 'col1'},
                    {'nullable': True,
                     'type': 'String',
                     'description': 'IHS',
                     'name': 'col2'},
                    {'nullable': False,
                     'type': 'String',
                     'description': 'IHS',
                     'name': 'col3'},
                  ]
            },
            'id': '2d113c3a-be1a-4c70-b87a-8e569dee8dd0'
        },
        'meta': {'total_count': 3}
    }


from dli.models import AttributesDict

class OrganisationModel(AttributesDict):
    @classmethod
    def _from_v2_response(cls, data):
        data = data.pop('data')
        attributes = data["attributes"]
        return cls(id=data["id"], **attributes)



from functools import partial
from unittest.mock import MagicMock
from tests.models.conftest import mock_catalogue_filter_search_response_impl


class TestPackageModel:
    def test_display_package(
        self, capsys, dataset_request_index_v2, test_client,
        package_request_v2
    ):
        test_client._session.get.return_value.json.side_effect = [
            dataset_request_index_v2
        ]

        package = test_client._Package(package_request_v2)
        package.contents()
        captured = capsys.readouterr()
        assert captured.out.startswith(
            '\nDATASET "TestDataset" [PARQUET]\n'
            '>> Shortcode: TestDataset\n'
            '>> Available Date Range: 2019-05-06 to 2020-01-01\n'
            '>> ID: 5b01376e-975d-11e9-8832-7e5ef76d533f\n'
            '>> Published: Monthly by IHS Markit\n'
        )

    def test_retrieve_datasets_given_package(self, test_client,
                                             package_request_index_v2, package_request_v2,
                                             dataset_request_index_v2):
        test_client._session.get.return_value.json.side_effect = [
            package_request_index_v2, dataset_request_index_v2
        ]

        pp = test_client.packages(search_term="", only_mine=False)
        ds = pp["Test Package"].datasets()
        assert len(ds) == 2


class TestPackageModule:

    def test_find_packages(
        self, test_client, package_request_index_v2
    ):


        test_client._session.get = partial(
            mock_catalogue_filter_search_response_impl,
            package_request_index_v2
        )

        assert len(test_client.packages()) == 1
        assert len(test_client.packages(search_term=[])) == 1
        assert len(test_client.packages(only_mine=False)) == 2
        assert len(test_client.packages(only_mine=True)) == 1
        assert len(test_client.packages(search_term=None, only_mine=False)) == 2
        assert len(test_client.packages(search_term=[], only_mine=False)) == 2
        assert len(test_client.packages(search_term=None, only_mine=True)) == 1
        assert len(test_client.packages(search_term=[], only_mine=True)) == 1

        # bad data is ignored (we filter that out before the send)
        assert len(test_client.packages(
            search_term=["baddata"], only_mine=False)) == 2
        assert len(test_client.packages(
            search_term=["baddata", "baddata"], only_mine=False)) == 2
        assert len(test_client.packages(
            search_term=["name=Test Package", "baddata"], only_mine=False)) == 1
        assert len(test_client.packages(
            search_term="baddata", only_mine=False)) == 2

        # testing search term with only_mine param against valid search term
        assert len(test_client.packages(
            search_term=["name=Test Package"], only_mine=False)) == 1
        assert len(test_client.packages(
            search_term=[], only_mine=True)) == 1
        assert len(test_client.packages(
            search_term=["name=Not a Package"], only_mine=True)) == 0

        # multi-param against only_mine
        assert len(test_client.packages(
            search_term=["name=Different Name", "topic=Finance",
                         "organisation_name=Test Org"],
            only_mine=True)) == 0
        assert len(test_client.packages(
            search_term=["name=Different Name", "topic=Finance",
                         "organisation_name=Test Org"],
            only_mine=False)) == 1
        assert len(test_client.packages(
            search_term=["name=Test Package", "topic=Climate",
                         "organisation_name=IHS Markit"],
            only_mine=False)) == 1
        assert len(test_client.packages(
            search_term=["description=test-data"],
            only_mine=False)) == 2

        # multi-param with bad data
        assert len(test_client.packages(
            search_term=["description=test-data", "organisation_name=baddata"],
            only_mine=False)) == 0

        # 'like' test
        assert len(test_client.packages(
            search_term=["name like Test"],
            only_mine=True)) == 1

        # test override of param against kwarg of same (accomplish both ways)
        assert len(test_client.packages(
            search_term=["has_access=True"],
            only_mine=False)) == 0

    def test_retrieve_package_by_get(
        self, test_client, package_request_index_v2, dataset_request_index_v2
    ):

        test_client._session.get.return_value.json.side_effect = [
            package_request_index_v2
        ]

        test_package = test_client.packages.get('Test Package')
        assert test_package.name == "Test Package"
